from copy import deepcopy

class Template_OBO:

    base_template = dict(
        pid=None,
        role_type=None,
        process_type=None,
        attribute_type=None,
        organisation=None,
        input_type=None,
        input_id=None,
        target_type=None,
        target_id=None,
        output_type=None,
        output_id=None,
        output_id_type=None,
        unit_type=None,
        specific_method_type=None,
        protocol_type=None,
        protocol_id=None,
        cause_type=None,
        cause_id=None,
        frequency_type=None,
        frequency_value=None,
        value_date=None,
        value_integer=None,
        value_string=None,
        value_id_string=None,
        value_float=None,
        value_datatype=None,
        comments=None,
        startdate=None,
        enddate=None,
        age=None,
        uniqid=None,
        event_id=None,
        value=None,
        valueIRI=None,
        activity=None,
        target=None,
        agent=None,
        input=None,
        unit=None,
    )

    @classmethod
    def build_entry(cls, **overrides):
        entry = deepcopy(cls.base_template)
        entry.update(overrides)
        return entry

TEMPLATE_MAP_OBO = {

        "Birthdate": Template_OBO.build_entry(
            protocol_type="http://purl.obolibrary.org/obo/OBI_0000272",
            role_type="http://purl.obolibrary.org/obo/OBI_0000093",
            process_type="http://purl.obolibrary.org/obo/NCIT_C142470",
            attribute_type="http://purl.obolibrary.org/obo/NCIT_C68615",
            output_type="http://purl.obolibrary.org/obo/NCIT_C70856",
            value_datatype="xsd:date"
        ),
        "Birthyear": Template_OBO.build_entry(
            protocol_type="http://purl.obolibrary.org/obo/OBI_0000272",
            role_type="http://purl.obolibrary.org/obo/OBI_0000093",
            process_type="http://purl.obolibrary.org/obo/NCIT_C142470",
            attribute_type="http://purl.obolibrary.org/obo/NCIT_C83164",
            output_type="http://purl.obolibrary.org/obo/NCIT_C70856",
            value_datatype="xsd:integer"
        ),
        "Birthplace": Template_OBO.build_entry(
            protocol_type="http://purl.obolibrary.org/obo/OBI_0000272",
            role_type="http://purl.obolibrary.org/obo/OBI_0000093",
            process_type="http://purl.obolibrary.org/obo/NCIT_C142470",
            attribute_type="http://purl.obolibrary.org/obo/NCIT_C176764",
            output_type="http://purl.obolibrary.org/obo/NCIT_C25464",
            output_id_type="http://purl.obolibrary.org/obo/NCIT_C20108",
            value_datatype="xsd:string"
        ),
        "Deathdate": Template_OBO.build_entry(
            protocol_type="http://purl.obolibrary.org/obo/OBI_0000272",
            role_type="http://purl.obolibrary.org/obo/OBI_0000093",
            process_type="http://purl.obolibrary.org/obo/NCIT_C142470",
            attribute_type="http://purl.obolibrary.org/obo/NCIT_C70810",
            output_type="http://purl.obolibrary.org/obo/NCIT_C70856",
            cause_type="http://purl.obolibrary.org/obo/NCIT_C81239",
            value_datatype="xsd:date"
        ),
        "First_visit": Template_OBO.build_entry(
            protocol_type="http://purl.obolibrary.org/obo/OBI_0000272",
            role_type="http://purl.obolibrary.org/obo/OBI_0000093",
            process_type="http://purl.obolibrary.org/obo/NCIT_C142470",
            attribute_type="http://purl.obolibrary.org/obo/NCIT_C159705",
            output_type="http://purl.obolibrary.org/obo/NCIT_C70856",
            value_datatype="xsd:date"
        ),
        "Symptoms_onset": Template_OBO.build_entry(
            protocol_type="http://purl.obolibrary.org/obo/OBI_0000272",
            role_type="http://purl.obolibrary.org/obo/OBI_0000093",
            process_type="http://purl.obolibrary.org/obo/NCIT_C142470",
            attribute_type="http://purl.obolibrary.org/obo/NCIT_C124353",
            output_type="http://purl.obolibrary.org/obo/NCIT_C70856",
            target_type="http://purl.obolibrary.org/obo/NCIT_C4876",
            value_datatype="xsd:date"
        ),
        "Sex": Template_OBO.build_entry(
            protocol_type="http://purl.obolibrary.org/obo/OBI_0000272",
            role_type="http://purl.obolibrary.org/obo/OBI_0000093",
            process_type="http://purl.obolibrary.org/obo/NCIT_C142470",
            output_type="http://purl.obolibrary.org/obo/NCIT_C160908",
            value_datatype="xsd:string"
        ),
        "Status": Template_OBO.build_entry(
            protocol_type="http://purl.obolibrary.org/obo/OBI_0000272",
            role_type="http://purl.obolibrary.org/obo/OBI_0000093",
            process_type="http://purl.obolibrary.org/obo/NCIT_C142470",
            output_type="http://purl.obolibrary.org/obo/OPMI_0000326",
            value_datatype="xsd:string"
        ),
        "Diagnosis": Template_OBO.build_entry(
            protocol_type="http://purl.obolibrary.org/obo/OBI_0000272",
            role_type="http://purl.obolibrary.org/obo/OBI_0000093",
            process_type="http://purl.obolibrary.org/obo/NCIT_C18020",
            attribute_type="http://purl.obolibrary.org/obo/NCIT_C7057",
            output_type="http://purl.obolibrary.org/obo/OGMS_0000073",
            output_id_type="http://purl.obolibrary.org/obo/NCIT_C154625",
            value_datatype="xsd:string"
        ),
        "Phenotype": Template_OBO.build_entry(
            protocol_type="http://purl.obolibrary.org/obo/OBI_0000272",
            role_type="http://purl.obolibrary.org/obo/OBI_0000093",
            process_type="http://purl.obolibrary.org/obo/NCIT_C18020",
            output_type="http://purl.obolibrary.org/obo/NCIT_C16977",
            output_id_type="http://purl.obolibrary.org/obo/NCIT_C164535",
            value_datatype="xsd:string"
        ),
        "Genetic": Template_OBO.build_entry(
            protocol_type="http://purl.obolibrary.org/obo/OBI_0000272",
            role_type="http://purl.obolibrary.org/obo/OBI_0000093",
            process_type="http://purl.obolibrary.org/obo/NCIT_C15709",
            output_type="http://purl.obolibrary.org/obo/NCIT_C171178",
            output_id_type="http://purl.obolibrary.org/obo/NCIT_C164607",
            value_datatype="xsd:string"
        ),
        "Examination": Template_OBO.build_entry(
            protocol_type="http://purl.obolibrary.org/obo/OBI_0000272",
            role_type="http://purl.obolibrary.org/obo/OBI_0000093",
            process_type="http://purl.obolibrary.org/obo/MAXO_0000487",
            output_type="http://purl.obolibrary.org/obo/NCIT_C70856",
            value_datatype="xsd:float"
        ),
        "Laboratory": Template_OBO.build_entry(
            protocol_type="http://purl.obolibrary.org/obo/OBI_0000272",
            role_type="http://purl.obolibrary.org/obo/OBI_0000093",
            process_type="http://purl.obolibrary.org/obo/NCIT_C25294",
            output_type="http://purl.obolibrary.org/obo/NCIT_C70856",
            value_datatype="xsd:float"
        ),
        "Surgery": Template_OBO.build_entry(
            protocol_type="http://purl.obolibrary.org/obo/OBI_0000272",
            role_type="http://purl.obolibrary.org/obo/OBI_0000093",
            process_type = "http://purl.obolibrary.org/obo/NCIT_C15329",
            value_datatype = "xsd:string"
        ),
        "Hospitalization": Template_OBO.build_entry(
            protocol_type="http://purl.obolibrary.org/obo/OBI_0000272",
            role_type="http://purl.obolibrary.org/obo/OBI_0000093",
            process_type = "http://purl.obolibrary.org/obo/NCIT_C25179",
            value_datatype = "xsd:string"
        ),
        "Medication": Template_OBO.build_entry(
            protocol_type="http://purl.obolibrary.org/obo/IAO_0000104",
            role_type="http://purl.obolibrary.org/obo/OBI_0000093",
            process_type = "http://purl.obolibrary.org/obo/NCIT_C70962",
            input_type = "http://purl.obolibrary.org/obo/NCIT_C177929",
            value_datatype = "xsd:float"
        ),
        "Clinical_trial": Template_OBO.build_entry(
            protocol_type="http://purl.obolibrary.org/obo/OBI_0000272",
            role_type="http://purl.obolibrary.org/obo/OBI_0000097",
            process_type = "http://purl.obolibrary.org/obo/NCIT_C71104",
            output_type = "http://purl.obolibrary.org/obo/NCIT_C142439",
            output_id_type="http://purl.obolibrary.org/obo/NCIT_C83082",
            target_type = "http://purl.obolibrary.org/obo/NCIT_C7057",
            value_datatype = "xsd:string"
        ),
        "Cohort": Template_OBO.build_entry(
            protocol_type="http://purl.obolibrary.org/obo/OBI_0000272",
            role_type="http://purl.obolibrary.org/obo/OBI_0000097",
            process_type = "http://purl.obolibrary.org/obo/NCIT_C15208",
            output_type = "http://purl.obolibrary.org/obo/NCIT_C142439",
            output_id_type="http://purl.obolibrary.org/obo/NCIT_C83082",
            target_type = "http://purl.obolibrary.org/obo/NCIT_C7057",
            value_datatype = "xsd:string"
        ),
        "Biobank": Template_OBO.build_entry(
            protocol_type="http://purl.obolibrary.org/obo/OBI_0000272",
            role_type="http://purl.obolibrary.org/obo/OBI_0000093",
            process_type = "http://purl.obolibrary.org/obo/OBIB_0000668",
            output_type = "http://purl.obolibrary.org/obo/NCIT_C19697", 
            output_id_type = "http://purl.obolibrary.org/obo/NCIT_C25402", 
            value_datatype = "xsd:string"
        ),
        "Consent": Template_OBO.build_entry(
            protocol_type="http://purl.obolibrary.org/obo/OBI_0000272",
            role_type="http://purl.obolibrary.org/obo/OBI_0000093",
            process_type = "http://purl.obolibrary.org/obo/OBI_0000810",
            attribute_type = "http://purl.obolibrary.org/obo/NCIT_C25460", 
            value_datatype = "xsd:string"
        ),
        "Questionnaire": Template_OBO.build_entry(
            role_type="http://purl.obolibrary.org/obo/OBI_0000093",
            process_type="http://purl.obolibrary.org/obo/NCIT_C20993",
            output_type="http://purl.obolibrary.org/obo/NCIT_C49149",
            input_type="http://purl.obolibrary.org/obo/NCIT_C91102",
            protocol_type="http://purl.obolibrary.org/obo/NCIT_C177377",
            value_datatype="xsd:float"
        ),
        "Disability": Template_OBO.build_entry(
            protocol_type="http://purl.obolibrary.org/obo/OBI_0000272",
            role_type="http://purl.obolibrary.org/obo/OBI_0000093",
            process_type="http://purl.obolibrary.org/obo/OMIT_0005448",
            attribute_type="http://purl.obolibrary.org/obo/NCIT_C21007",
            output_type="http://purl.obolibrary.org/obo/NCIT_C49149",
            value_datatype="xsd:float"
        ),
    }

class Template_SNOMED:

    base_template = dict(
        pid=None,
        role_type="http://snomed.info/id/116154003",
        process_type=None,
        attribute_type=None,
        organisation=None,
        input_type=None,
        input_id=None,
        target_type=None,
        target_id=None,
        output_type=None,
        output_id=None,
        output_id_type=None,
        unit_type=None,
        specific_method_type=None,
        protocol_type=None,
        protocol_id=None,
        cause_type=None,
        cause_id=None,
        frequency_type=None,
        frequency_value=None,
        value_date=None,
        value_integer=None,
        value_string=None,
        value_id_string=None,
        value_float=None,
        value_datatype=None,
        comments=None,
        startdate=None,
        enddate=None,
        age=None,
        uniqid=None,
        event_id=None,
        value=None,
        valueIRI=None,
        activity=None,
        target=None,
        agent=None,
        input=None,
        unit=None,
    )

    @classmethod
    def build_entry(cls, **overrides):
        entry = deepcopy(cls.base_template)
        entry.update(overrides)
        return entry

TEMPLATE_MAP_SNOMED = {
        "Birthdate":Template_SNOMED.build_entry(
            role_type="http://snomed.info/id/116154003",
            protocol_type="http://snomed.info/id/258049002",
            process_type="http://snomed.info/id/312486000",
            attribute_type="http://snomed.info/id/3950001",
            output_type="http://snomed.info/id/184099003",
            value_datatype="xsd:date"
        ),
        "Birthyear":Template_SNOMED.build_entry(
            role_type="http://snomed.info/id/116154003",
            protocol_type="http://snomed.info/id/258049002",
            process_type="http://snomed.info/id/312486000",
            attribute_type="http://snomed.info/id/3950001",
            output_type="http://snomed.info/id/258707000",
            value_datatype="xsd:integer"
        ),
        "Birthplace":Template_SNOMED.build_entry(
            role_type="http://snomed.info/id/116154003",
            protocol_type="http://snomed.info/id/258049002",
            process_type="http://snomed.info/id/312486000",
            output_type="http://snomed.info/id/315354004",
            output_id_type="http://snomed.info/id/118522005",
            value_datatype="xsd:integer"
        ),
        "Deathdate":Template_SNOMED.build_entry(
            role_type="http://snomed.info/id/116154003",
            protocol_type="http://snomed.info/id/258049002",
            process_type="http://snomed.info/id/363049002",
            attribute_type="http://snomed.info/id/419620001",
            output_type="http://snomed.info/id/399753006",
            cause_type="http://snomed.info/id/184305005",
            value_datatype="xsd:date"
        ),
        "First_visit":Template_SNOMED.build_entry(
            role_type="http://snomed.info/id/116154003",
            protocol_type="http://snomed.info/id/258049002",
            process_type="http://snomed.info/id/308335008",
            attribute_type="http://snomed.info/id/769681006",
            output_type="http://snomed.info/id/406543005",
            value_datatype="xsd:date"
        ),
        "Symptoms_onset":Template_SNOMED.build_entry(
            role_type="http://snomed.info/id/116154003",
            protocol_type="http://snomed.info/id/258049002",
            process_type="http://snomed.info/id/308335008",
            attribute_type="http://snomed.info/id/308918001",
            output_type="http://snomed.info/id/263501003",
            target_type="http://snomed.info/id/162408000",
            value_datatype="xsd:date"
        ),
        "Sex":Template_SNOMED.build_entry(
            role_type="http://snomed.info/id/116154003",
            protocol_type="http://snomed.info/id/258049002",
            process_type="http://snomed.info/id/312486000",
            output_type="http://snomed.info/id/734000001",
            value_datatype="xsd:string"
        ),
        "Status":Template_SNOMED.build_entry(
            role_type="http://snomed.info/id/116154003",
            protocol_type="http://snomed.info/id/258049002",
            process_type="http://snomed.info/id/386053000",
            output_type="http://snomed.info/id/420107008",
            value_datatype="xsd:string"
        ),
        "Diagnosis":Template_SNOMED.build_entry(
            role_type="http://snomed.info/id/116154003",
            protocol_type="http://snomed.info/id/258049002",
            process_type="http://snomed.info/id/103693007",
            output_type="http://snomed.info/id/439401001",
            output_id_type="http://snomed.info/id/118522005",
            attribute_type="http://snomed.info/id/64572001",
            value_datatype="xsd:string"
        ),
        "Phenotype":Template_SNOMED.build_entry(
            role_type="http://snomed.info/id/116154003",
            protocol_type="http://snomed.info/id/258049002",
            process_type="http://snomed.info/id/363778006",
            output_type="http://snomed.info/id/8116006",
            output_id_type="http://snomed.info/id/118522005",
            value_datatype="xsd:string"
        ),
        "Genetic":Template_SNOMED.build_entry(
            role_type="http://snomed.info/id/116154003",
            protocol_type="http://snomed.info/id/258049002",
            process_type="http://snomed.info/id/405824009",
            output_type="http://snomed.info/id/41482005",
            output_id_type="http://snomed.info/id/118522005",
            value_datatype="xsd:string"
        ),
        "Questionnaire":Template_SNOMED.build_entry(
            role_type="http://snomed.info/id/116154003",
            protocol_type ="http://snomed.info/id/1402981000000101",
            process_type="http://snomed.info/id/840297006",
            input_type = "https://loinc.org/LP175698-2",
            output_type = "https://loinc.org/82783-2",
            value_datatype="xsd:float"
        ),
        "Disability":Template_SNOMED.build_entry(
            role_type="http://snomed.info/id/116154003",
            protocol_type="http://snomed.info/id/258049002",
            process_type="http://snomed.info/id/81078003",
            attribute_type="http://snomed.info/id/21134002",
            output_type="http://snomed.info/id/273421001",
            value_datatype="xsd:float"
        ),
        "Examination":Template_SNOMED.build_entry(
            role_type="http://snomed.info/id/116154003",
            protocol_type="http://snomed.info/id/258049002",
            process_type="http://snomed.info/id/315306007",
            output_type="http://snomed.info/id/363789004",
            value_datatype="xsd:float"
        ),
        "Laboratory":Template_SNOMED.build_entry(
            role_type="http://snomed.info/id/116154003",
            protocol_type="http://snomed.info/id/258049002",
            process_type="http://snomed.info/id/108252007",
            output_type="http://snomed.info/id/4241000179101",
            value_datatype="xsd:float"
        ),
        "Surgery":Template_SNOMED.build_entry(
            role_type="http://snomed.info/id/116154003",
            protocol_type="http://snomed.info/id/258049002",
            process_type = "http://snomed.info/id/387713003",
            value_datatype = "xsd:string"
        ),
        "Hospitalization":Template_SNOMED.build_entry(
            role_type="http://snomed.info/id/116154003",
            protocol_type="http://snomed.info/id/258049002",
            process_type = "https://loinc.org/LA15417-1",
            value_datatype = "xsd:string"
        ),
        "Medication":Template_SNOMED.build_entry(
            role_type="http://snomed.info/id/116154003",
            protocol_type="http://snomed.info/id/258049002",
            process_type = "http://snomed.info/id/18629005",
            input_type = "http://snomed.info/id/2533004",      
            output_type = "https://loinc.org/18615-5",
            value_datatype = "xsd:float"
        ),
        "Clinical_trial":Template_SNOMED.build_entry(
            protocol_type="http://snomed.info/id/258049002",
            role_type="http://snomed.info/id/428024001",
            process_type = "http://snomed.info/id/110465008",
            output_id_type = "http://snomed.info/id/118522005",
            output_type = "http://snomed.info/id/229059009", 
            target_type ="http://snomed.info/id/64572001", 
            value_datatype = "xsd:string"
        ),
        "Cohort":Template_SNOMED.build_entry(
            protocol_type="http://snomed.info/id/258049002",
            role_type="http://snomed.info/id/428024001",
            process_type = "http://snomed.info/id/719581000000108",
            output_id_type = "http://snomed.info/id/118522005",
            output_type = "http://snomed.info/id/229059009", 
            target_type ="http://snomed.info/id/64572001", 
            value_datatype = "xsd:string"
        ),
        "Biobank":Template_SNOMED.build_entry(
            role_type="http://snomed.info/id/116154003",
            protocol_type="http://snomed.info/id/258049002",
            process_type = "http://snomed.info/id/433465004",
            output_type = "http://snomed.info/id/364611000000101", 
            output_id_type = "http://snomed.info/id/118522005",     
            value_datatype = "xsd:string"
        )
    }


import pandas as pd
import os
import logging
from datetime import datetime
from .template import TEMPLATE_MAP_OBO, TEMPLATE_MAP_SNOMED

logging.basicConfig(level=logging.INFO, force=True)


class Toolkit:
    # GLOBAL REMOVAL COUNTER
    removed_rows_total = 0

    keywords_OBO = {
        "Birthdate", "Birthyear", "Deathdate", "Sex", "Birthplace", "Status", "First_visit",
        "Questionnaire", "Diagnosis", "Phenotype", "Symptoms_onset", "Examination", "Laboratory",
        "Genetic", "Disability", "Medication", "Surgery", "Hospitalization",
        "Consent", "Biobank", "Clinical_trial", "Cohort"
    }

    keywords_SNOMED = {
        "Birthdate", "Birthyear", "Deathdate", "Sex", "Birthplace", "Status", "First_visit",
        "Questionnaire", "Diagnosis", "Phenotype", "Symptoms_onset", "Examination", "Laboratory",
        "Genetic", "Disability", "Medication", "Surgery", "Hospitalization",
        "Biobank", "Clinical_trial", "Cohort"
    }

    columns = [
        "model", "pid", "event_id", "value", "age", "value_datatype", "valueIRI", "activity",
        "unit", "input", "target", "protocol_id", "specification", "frequency_type",
        "frequency_value", "agent", "startdate", "enddate", "comments", "organisation"
    ]

    drop_columns = ["value", "valueIRI", "target", "agent", "input", "activity", "unit"]

    columns_to_check = [
        "valueIRI", "activity", "unit", "input", "target",
        "frequency_type", "agent", "specification", "organisation"
    ]

    # UTILITIES
    @staticmethod
    def milisec():
        return datetime.now().strftime('%Y%m%d%H%M%S%f')

    def get_template(self, template_type):
        if template_type == "OBO":
            return TEMPLATE_MAP_OBO
        elif template_type == "SNOMED":
            return TEMPLATE_MAP_SNOMED
        raise ValueError(f"Template type '{template_type}' not recognized.")

    # CENTRALIZED REMOVAL LOGGER
    def _mark_row_for_removal(self, filepath, idx, reason, columns=None):
        self.removed_rows_total += 1
        col_info = f" | columns: {columns}" if columns else ""
        logging.info(
            f"[{os.path.basename(filepath)}] "
            f"Row {idx} removed | reason: {reason}{col_info}"
        )

    # MAIN PIPELINE
    def whole_method(self, folder_path, template_type):
        matching_files = self._find_matching_files(folder_path, template_type)
        processed = [self._process_file(file, template_type) for file in matching_files]
        final_df = pd.concat(processed, ignore_index=True) if processed else pd.DataFrame(columns=self.columns)
        final_df = self.delete_extra_columns(final_df)
        final_df.to_csv(os.path.join(folder_path, "CARE.csv"), index=False)

    def _find_matching_files(self, folder_path, template_type):
        keywords = self.keywords_OBO if template_type == "OBO" else self.keywords_SNOMED
        return [
            os.path.join(folder_path, file)
            for file in os.listdir(folder_path)
            if file.endswith(".csv") and any(k in file for k in keywords)
        ]

    def _process_file(self, filepath, template_type):
        df = self.import_your_data_from_csv(filepath)
        if df is None:
            return pd.DataFrame(columns=self.columns)

        df = self.check_status_column_names(df)
        df = self.add_columns_from_template(df, template_type)
        df = self.iri_validation(df, filepath)
        df = self.value_edition(df)
        df = self.time_edition(df)
        df = self.clean_empty_rows(df, filepath)
        df = self.unique_id_generation(df)
        df = self.normalize_numeric_columns(df)

        logging.info(f"Transformed: {os.path.basename(filepath)}")
        logging.info(f"TOTAL ROWS REMOVED ACROSS ALL FILES: {self.removed_rows_total}\n")
        self.removed_rows_total = 0

        return df

    # CSV HANDLING
    def import_your_data_from_csv(self, filepath):
        try:
            df = pd.read_csv(
                filepath,
                dtype={c: str for c in self.columns_to_check if c in pd.read_csv(filepath, nrows=0).columns}
            )
            logging.info(f"Imported CSV: {os.path.basename(filepath)}")
            return df
        except Exception as e:
            logging.error(f"Error loading CSV {filepath}: {e}")
            return None

    def check_status_column_names(self, df):
        extra = set(df.columns) - set(self.columns)
        if extra:
            raise ValueError(f"Unexpected columns: {extra}")
        for col in self.columns:
            if col not in df.columns:
                df[col] = pd.Series(dtype=object)
        return df

    def add_columns_from_template(self, df, template_type):
        template = self.get_template(template_type)
        rows = []
        for _, row in df.iterrows():
            base = {"model": row["model"]}
            base.update(template.get(row["model"], {}))
            base.update({k: v for k, v in row.items() if pd.notnull(v)})
            rows.append(base)
        return pd.DataFrame(rows)

    # IRI VALIDATION AND ROW REMOVAL
    def iri_validation(self, df, filepath):
        def is_valid_iri(v):
            return isinstance(v, str) and (v.startswith("http://") or v.startswith("https://"))

        rows_to_drop = []

        for idx, row in df.iterrows():
            for col in self.columns_to_check:
                val = row.get(col)
                if pd.isna(val) or str(val).strip() == "":
                    continue
                if not is_valid_iri(str(val).strip()):
                    self._mark_row_for_removal(
                        filepath, idx,
                        "Non-IRI value detected",
                        {col: val}
                    )
                    rows_to_drop.append(idx)
                    break

        return df.drop(index=rows_to_drop).reset_index(drop=True)
        
    def value_edition(self, df):
        def apply_value_types(row):
            model = row.get("model")
            val = row.get("value")
            dtype = row.get("value_datatype")

            # ---------- Literal values ----------
            if pd.notnull(val):
                if dtype == "xsd:string":
                    if model in [
                        "Diagnosis", "Phenotype", "Birthplace", "Genetic",
                        "Clinical_trial", "Cohort", "Biobank"
                    ]:
                        row["value_id_string"] = val
                    else:
                        row["value_string"] = val

                elif dtype == "xsd:float":
                    row["value_float"] = val

                elif dtype == "xsd:integer":
                    row["value_integer"] = val

                elif dtype == "xsd:date":
                    row["value_date"] = val

            # ---------- Dispatch maps ----------
            expected_valueIRI_models = {
                "attribute_type": ["Sex", "Status", "Examination"],
                "output_id": ["Diagnosis", "Phenotype", "Birthplace", "Genetic", "Clinical_trial", "Cohort", "Biobank"],
                "cause_id": ["Deathdate"],
                "output_type": ["Consent", "Medication"],
            }

            expected_target_models = {
                "target_id": ["Symptoms_onset", "Clinical_trial", "Cohort"],
                "target_type": ["Examination", "Laboratory", "Surgery", "Diagnosis", "Phenotype"],
            }

            expected_input_models = {
                "input_type": ["Laboratory", "Genetic", "Biobank"],
                "input_id": ["Questionnaire", "Disability"],
            }

            expected_agent_models = {
                "attribute_type": ["Genetic"],
                "input_id": ["Medication"],
            }

            expected_activity_models = {
                "specific_method_type": [
                    "Examination", "Laboratory", "Genetic",
                    "Medication", "Hospitalization", "Surgery"
                ]
            }

            expected_unit_models = {
                "unit_type": self.keywords_OBO
            }

            # ---------- Dispatcher ----------
            def dispatch(column, value, expected_map):
                if pd.isnull(value):
                    return

                for new_col, valid_models in expected_map.items():
                    if model in valid_models:
                        row[new_col] = value
                        return

                logging.info(
                    f"Unexpected value in '{column}' for model '{model}': {value}"
                )

            # ---------- Apply dispatchers ----------
            dispatch("valueIRI", row.get("valueIRI"), expected_valueIRI_models)
            dispatch("target", row.get("target"), expected_target_models)
            dispatch("input", row.get("input"), expected_input_models)
            dispatch("agent", row.get("agent"), expected_agent_models)
            dispatch("activity", row.get("activity"), expected_activity_models)
            dispatch("unit", row.get("unit"), expected_unit_models)

            return row

        return df.apply(apply_value_types, axis=1)


    # EMPTY ROWS AND ROW REMOVAL
    def clean_empty_rows(self, df, filepath):
        required = ["value", "valueIRI", "activity", "target", "agent"]
        rows_to_drop = []

        for idx, row in df.iterrows():
            if row[required].isnull().all():
                self._mark_row_for_removal(
                    filepath, idx,
                    "Insufficient data (all required columns empty)"
                )
                rows_to_drop.append(idx)

        return df.drop(index=rows_to_drop).reset_index(drop=True)

    # FINAL STEPS
    def time_edition(self, df):
        if "startdate" in df and "enddate" in df:
            df["enddate"] = df["enddate"].where(df["enddate"].notna(), df["startdate"])
        return df

    def delete_extra_columns(self, df):
        return df.drop(columns=[c for c in self.drop_columns if c in df.columns], errors="ignore")

    def unique_id_generation(self, df):
        ts = self.milisec()
        df["uniqid"] = [f"{ts}{i}" for i in range(len(df))]
        return df
    
    def normalize_numeric_columns(self, df):
        """
        Force integer-like columns to use pandas nullable Int64 dtype
        to avoid 30 -> 30.0 caused by NaN values.
        """
        int_columns = [
            "age",
            "value_integer",
            "frequency_value",
            "pid",
            "event_id"
        ]

        for col in int_columns:
            if col in df.columns:
                df[col] = pd.to_numeric(df[col], errors="coerce").astype("Int64")

        return df